@@ -221,6 +221,7 @@ class Metal(object):
 	An object for paramagnetic chi tensors and delta-chi tensors.
 	This class has basic attributes that specify position, 
 	axiality/rhombicity, isotropy and euler angles.
+	It also has methods for calculating PCS, RDC, PRE and CCR values.
 	"""
 
 	# Gyromagnetic ratio of an electron and other constants
@@ -231,6 +232,7 @@ class Metal(object):
 	GAMMA = 1.760859644E11
 
 	# Stored values get scaled by this amount for the fitting algorithm
+	# This ensures against floating point errors during least-squares
 	fit_scaling = {
 		'x': 1E10,
 		'y': 1E10,
@@ -244,17 +246,20 @@ class Metal(object):
 		't1e': 1E12,
 		'taur': 1E9,
 		'mueff': 1./MUB,
+		'gax': 1E60,
+		'grh': 1E60
 	}
 
 	# J, g, T1e values for lanthanide series
-	lanth_lib = OrderedDict([('Zero',(0., 0. , 0.)),
+	lanth_lib = OrderedDict([
+		('Zero',(0.   , 0.    , 0.       )),
 		('Ce', ( 5./2., 6./7. , 0.133E-12)),
 		('Pr', ( 4./1., 4./5. , 0.054E-12)),
 		('Nd', ( 9./2., 8./11., 0.210E-12)),
-		('Pm', ( 4./1., 3./5. , np.nan   )),
+		('Pm', ( 4./1., 3./5. , 0.0      )),
 		('Sm', ( 5./2., 2./7. , 0.074E-12)),
 		('Eu', ( 2./1., 3./2. , 0.015E-12)),
-		('Gd', ( 7./2., 2./1. , np.nan   )),
+		('Gd', ( 7./2., 2./1. , 0.0      )),
 		('Tb', ( 6./1., 3./2. , 0.251E-12)),
 		('Dy', (15./2., 4./3. , 0.240E-12)),
 		('Ho', ( 8./1., 5./4. , 0.209E-12)),
@@ -263,8 +268,9 @@ class Metal(object):
 		('Yb', ( 7./2., 8./7. , 0.157E-12))]
 	)
 
-	# Template anisotropies [axial, rhombic]
-	lanth_axrh = OrderedDict([('Zero',(0. , 0.)),
+	# Template anisotropies [axial, rhombic] from Bertini
+	lanth_axrh = OrderedDict([
+		('Zero',( 0.0,  0.0)),
 		('Ce', (  2.1,  0.7)),
 		('Pr', (  3.4,  2.1)),
 		('Nd', (  1.7,  0.4)),
@@ -280,12 +286,27 @@ class Metal(object):
 		('Yb', ( -8.3, -5.8))]
 	)
 
-	# Indices defining 5 unique elements of tensor anisotropy
+	fundamental_attributes = (
+		'position', 
+		'eulers', 
+		'axrh', 
+		'mueff', 
+		'g_axrh', 
+		't1e',
+		'shift', 
+		'temperature', 
+		'B0', 
+		'taur'
+		)
+
+	# Indices defining 5 unique elements of 3x3 tensor anisotropy
 	upper_coords = ((0,1,0,0,1),(0,1,1,2,2))
 	lower_coords = ((0,1,1,2,2),(0,1,0,0,1))
 
 	def __init__(self, position=(0,0,0), eulers=(0,0,0), 
-		axrh=(0,0), mueff=0.0, shift=0.0, temperature=298.15, t1e=0.0,
+		axrh=(0,0), mueff=0.0, 
+		g_axrh=(0,0), t1e=0.0,
+		shift=0.0, temperature=298.15, 
 		B0=18.79, taur=0.0):
 		"""
 		Instantiate ChiTensor object
@@ -301,6 +322,9 @@ class Metal(object):
 		axrh : array of floats, optional
 			the axial and rhombic values defining the magnetic susceptibility
 			anisotropy
+		g_axrh : array of floats, optional
+			the axial and rhombic values defining the power spectral density
+			tensor
 		mueff : float
 			the effective magnetic moment in units of A.m^2
 		shift : float
@@ -319,6 +343,7 @@ class Metal(object):
 		self.position = np.array(position, dtype=float)
 		self.eulers = np.array(eulers, dtype=float)
 		self.axrh = np.array(axrh, dtype=float)
+		self.g_axrh = np.array(g_axrh, dtype=float)
 		self.mueff = mueff
 		self.shift = shift
 		self.temperature = temperature
@@ -326,6 +351,9 @@ class Metal(object):
 		self.B0 = B0
 		self.taur = taur
 
+		# A dictionary of volatile parameters used during fitting
+		self.par = {}
+
 	def __str__(self):
 		return str(self.tensor)
 
@@ -343,6 +371,8 @@ class Metal(object):
 		.. math::
 			\\tau_c = \\frac{1}{\\frac{1}{\\tau_r}+\\frac{1}{T_{1e}}}
 		"""
+		if self.taur==0.0 or self.t1e==0.0:
+			raise ValueError("You need to set both taur and t1e attributes")
 		return 1./(1./self.taur + 1./self.t1e)
 
 	def copy(self):
@@ -357,7 +387,8 @@ class Metal(object):
 		return self.__class__(
 			position=tuple(self.position), 
 			eulers=tuple(self.eulers),
-			axrh=tuple(self.axrh), 
+			axrh=tuple(self.axrh),
+			g_axrh=tuple(self.g_axrh),
 			mueff=self.mueff, 
 			shift=self.shift, 
 			temperature=self.temperature, 
@@ -373,8 +404,8 @@ class Metal(object):
 		Parameters
 		----------
 		lanthanide : str
-			one of ['Ce','Pr','Nd','Pm','Sm','Eu','Gd',
-			        'Tb','Dy','Ho','Er','Tm','Yb']
+			one of 'Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb',
+			'Dy','Ho','Er','Tm','Yb'
 		set_dichi : bool (optional)
 			if True (default), the tensor anisotropy is set.
 			Otherwise only the isotropy and T1e values are set 
@@ -591,6 +622,20 @@ class Metal(object):
 	@B0_MHz.setter
 	def B0_MHz(self, value):
 		self.B0 = value / 42.57747892
+	@property
+	def gax(self):
+		"""axial componenet of spectral power density tensor"""
+		return self.g_axrh[0]
+	@gax.setter
+	def gax(self, value):
+		self.g_axrh[0] = value
+	@property
+	def grh(self):
+		"""axial componenet of spectral power density tensor"""
+		return self.g_axrh[1]
+	@gax.setter
+	def grh(self, value):
+		self.g_axrh[1] = value
 
 	@property
 	def eigenvalues(self):
@@ -610,10 +655,34 @@ class Metal(object):
 
 	@isotropy.setter
 	def isotropy(self, newIsotropy):
-		if newIsotropy<0:
+		if newIsotropy<=0:
 			newIsotropy = 0.0
 		self.mueff = ((newIsotropy*3*self.K*self.temperature) / self.MU0)**0.5
 
+	@property
+	def g_eigenvalues(self):
+		"""The eigenvalues defining the magnitude of the principle axes"""
+		return anisotropy_to_eigenvalues(self.g_axrh) + self.g_isotropy
+
+	@g_eigenvalues.setter
+	def g_eigenvalues(self, newEigenvalues):
+		self.g_axrh = eigenvalues_to_anisotropy(newEigenvalues)
+		# self.isotropy = np.round(np.sum(newEigenvalues)/3., 40)
+		self.g_isotropy = np.sum(newEigenvalues)/3.
+
+	@property
+	def g_isotropy(self):
+		"""Estimate of the spectral power density tensor isotropy"""
+		return (self.t1e * self.isotropy * self.K * self.temperature) / self.MU0
+
+	@g_isotropy.setter
+	def g_isotropy(self, newIsotropy):
+		if newIsotropy<=0 or self.isotropy<=0:
+			self.t1e = 0.0
+		else:	
+			self.t1e = (newIsotropy * self.MU0) / (
+				self.isotropy * self.K * self.temperature)
+
 	@property
 	def rotationMatrix(self):
 		"""The rotation matrix as defined by the euler angles"""
@@ -722,6 +791,24 @@ class Metal(object):
 		newTensor[2,2] = - elements[0] - elements[1]
 		self.tensor_saupe = newTensor
 
+	@property
+	def g_tensor(self):
+		"""The magnetic susceptibility tensor matrix representation"""
+		R = self.rotationMatrix
+		return R.dot(np.diag(self.g_eigenvalues)).dot(R.T)
+
+	@g_tensor.setter
+	def g_tensor(self, newTensor):
+		eigenvals, eigenvecs = np.linalg.eigh(newTensor)
+		eigs = zip(eigenvals, np.array(eigenvecs).T)
+		iso = np.sum(eigenvals)/3.
+		eigenvals, (x, y, z) = zip(*sorted(eigs, key=lambda x: abs(x[0]-iso)))
+		eigenvecs = x * z.dot(np.cross(x,y)), y, z
+		rotationMatrix = np.vstack(eigenvecs).T
+		eulers = unique_eulers(matrix_to_euler(rotationMatrix))
+		self.eulers = np.array(eulers, dtype=float)
+		self.g_eigenvalues = eigenvals
+
 	def set_utr(self):
 		"""
 		Modify current tensor parameters to unique tensor representation (UTR)
@@ -744,6 +831,23 @@ class Metal(object):
 		"""
 		self.position = atom.position
 
+	def average(self, metals):
+		"""
+		Set the attributes of the current instance to the average
+		of a list of provided tensor objects
+
+		WARNING: averging is unstable for spectral power density <g_tensor>
+
+		Parameters
+		----------
+		metals : a list of Metal objects
+			the average of attributes of this list will be taken
+		"""
+		self.g_tensor = sum([m.g_tensor for m in metals])/len(metals)
+		self.tensor = sum([m.tensor for m in metals])/len(metals)
+		self.position = sum(m.position for m in metals)/len(metals)
+		self.taur = sum([m.taur for m in metals])/len(metals)
+
 	################################
 	# Methods for PCS calculations #
 	################################
@@ -1330,7 +1434,65 @@ class Metal(object):
 			13*self.spec_dens(self.tauc, self.B0*self.GAMMA))
 		return rate
 
-	def pre(self, position, gamma, rtype, dsa=True, sbm=True, csa=0.0):
+	def g_sbm_r1(self, position, gamma):
+		"""
+		Calculate R1 relaxation due to Solomon-Bloembergen-Morgan theory
+		from anisotropic power spectral density tensor
+
+		Parameters
+		----------
+		position : array of floats
+			three coordinates (x,y,z)
+		gamma : float
+			the gyromagnetic ratio of the spin
+
+		Returns
+		-------
+		value : float
+			The R1 relaxation rate in /s
+		"""
+		pos = np.array(position, dtype=float) - self.position
+		distance = np.linalg.norm(pos)
+		pos_unit = pos / distance
+		preFactor = (2./3.) * (self.MU0 / (4*np.pi))**2
+		preFactor *= gamma**2 / distance**6
+		D = 3*np.kron(pos_unit,pos_unit).reshape(3,3) - np.identity(3)
+		return preFactor * (D.dot(D)).dot(self.g_tensor).trace()
+
+	def fast_g_sbm_r1(self, posarray, gammaarray):
+		"""
+		Vectorised version of :meth:`paramagpy.metal.Metal.g_sbm_r1`
+
+		This is generally used for speed in fitting PRE data
+
+		Parameters
+		----------
+		posarray : array with shape (n,3)
+			array of positions in meters
+		gammaarray : array with shape (n,3)
+			array of gyromagnetic ratios of the spins
+
+		Returns
+		-------
+		rates : array with shape (n,1)
+			The R1 relaxation rates in /s
+		"""
+		n = len(gammaarray)
+		pos = posarray - self.position
+		distance = np.linalg.norm(pos, axis=1)
+		pos_unit = (pos.T / distance).T
+		preFactor = (2./3.) * (self.MU0 / (4*np.pi))**2
+		preFactor *= gammaarray**2 / distance**6
+		D1 = np.einsum('ij,ik->ijk', pos_unit, pos_unit)
+		D2 = np.tile(np.identity(3), n).T.reshape(n,3,3)
+		D = 3*D1 - D2
+		D_squared = np.einsum('ijk,ilk->ijl', D,D)
+		tmp = np.einsum('ijk,lk->ijl', D_squared, self.g_tensor)
+		return preFactor * np.einsum('ijj->i', tmp)
+
+
+	def pre(self, position, gamma, rtype, dsa=True, sbm=True, 
+			gsbm=False, csa=0.0):
 		"""
 		Calculate the PRE for a set of spins using Curie and or SBM theory
 
@@ -1345,7 +1507,12 @@ class Metal(object):
 		dsa : bool (optional)
 			when True (defualt), DSA or Curie spin relaxation is included
 		sbm : bool (optional)
-			when True (defualt), SBM spin relaxation is included 
+			when True (defualt), SBM spin relaxation is included
+		gsbm : bool (optional)
+			when True (default=False), anisotropic dipolar relaxation is 
+			included using the spectral power density gensor <g_tensor>
+			NOTE: when true, ignores relaxation of type SBM
+			NOTE: only implemented for R1 relaxation calculations
 		csa : array with shape (3,3) (optional)
 			CSA tensor of the spin.
 			This defualts to 0.0, meaning CSAxDSA crosscorrelation is
@@ -1356,12 +1523,19 @@ class Metal(object):
 		rate : float
 			The PRE rate in /s
 		"""
+		if gsbm:
+			sbm = False
+			if rtype=='r2':
+				raise NotImplementedError(
+					"Anisotropic dipolar relaxation has not been implement for R2 caluculations yet.")
 		rate = 0.0
 		if rtype=='r1':
 			if dsa:
 				rate += self.dsa_r1(position, gamma, csa)
 			if sbm:
 				rate += self.sbm_r1(position, gamma)
+			if gsbm:
+				rate += self.g_sbm_r1(position, gamma)
 		elif rtype=='r2':
 			if dsa:
 				rate += self.dsa_r2(position, gamma, csa)
@@ -1370,7 +1544,7 @@ class Metal(object):
 		return rate
 
 	def fast_pre(self, posarray, gammaarray, rtype, 
-		dsa=True, sbm=True, csaarray=0.0):
+		dsa=True, sbm=True, gsbm=False, csaarray=0.0):
 		"""
 		Calculate the PRE for a set of spins using Curie and or SBM theory
 
@@ -1385,7 +1559,12 @@ class Metal(object):
 		dsa : bool (optional)
 			when True (defualt), DSA or Curie spin relaxation is included
 		sbm : bool (optional)
-			when True (defualt), SBM spin relaxation is included 
+			when True (defualt), SBM spin relaxation is included
+		gsbm : bool (optional)
+			when True (default=False), anisotropic dipolar relaxation is 
+			included using the spectral power density gensor <g_tensor>
+			NOTE: when true, ignores relaxation of type SBM
+			NOTE: only implemented for R1 relaxation calculations
 		csaarray : array with shape (m,3,3) (optional)
 			array of CSA tensors of the spins.
 			This defualts to 0.0, meaning CSAxDSA crosscorrelation is
@@ -1396,12 +1575,16 @@ class Metal(object):
 		rates : array with shape (n,1)
 			The PRE rates in /s
 		"""
+		if gsbm:
+			sbm = False
 		rates = 0.0
 		if rtype=='r1':
 			if dsa:
 				rates += self.fast_dsa_r1(posarray, gammaarray, csaarray)
 			if sbm:
 				rates += self.fast_sbm_r1(posarray, gammaarray)
+			if gsbm:
+				rates += self.fast_g_sbm_r1(posarray, gammaarray)
 		elif rtype=='r2':
 			if dsa:
 				rates += self.fast_dsa_r2(posarray, gammaarray, csaarray)
@@ -1465,6 +1648,7 @@ class Metal(object):
 		value : float
 			The R2 differential line broadening rate in /s
 		"""
+		# NOTE: B0 factor expresses shielding tensor in ppm
 		shield = dipole_shift_tensor/self.B0
 		spinUp   = self.dsa_r2(position, gamma, csa= shield)
 		spinDown = self.dsa_r2(position, gamma, csa=-shield)
@@ -1494,6 +1678,7 @@ class Metal(object):
 		rates : array with shape (n,1)
 			The R2 differential line broadening rates in /s
 		"""
+		# NOTE: B0 factor expresses shielding tensor in ppm
 		shield = dstarray/self.B0
 		spinUp   = self.fast_dsa_r2(posarray, gammaarray, 
 			csaarray= shield)
@@ -1607,9 +1792,9 @@ class Metal(object):
 
 		Parameters
 		----------
-		density : int
+		density : int (optional)
 			the points per Angstrom in the grid
-		size : float
+		size : float (optional)
 			the length of one edge of the cube
 
 		Returns
@@ -1617,19 +1802,18 @@ class Metal(object):
 		mesh : cubic grid array
 			This has shape (n,n,n,3) where n is the number of points
 			along one edge of the grid. Units are meters
-		bounds : tuple (origin, low, high, points)
-			This tuple has information about the bounding box
-
-			origin : array of floats, the (x,y,z) location of mesh vertex
-
-			low : array of ints, the integer location of the first point in each dimension
-
-			high : array of ints, the integer location of the last point in each dimension
-
-			points : array of ints, the number of points along each dimension
+		origin : array of floats, 
+			the (x,y,z) location of mesh vertex
+		low : array of ints, the integer location of the first 
+			point in each dimension
+		high : array of ints, the integer location of the last 
+			point in each dimension
+		points : array of ints, 
+			the number of points along each dimension
 
 		"""
-		origin = np.asarray(density * (self.position*1E10 - size/2.0), dtype=int)
+		origin = np.asarray(density * (self.position*1E10 - size/2.0), 
+			dtype=int)
 		low = origin / float(density)
 		high = low + size
 		points = np.array([int(density*size)]*3) + 1
@@ -1655,8 +1839,55 @@ class Metal(object):
 		pcs_mesh = self.fast_pcs(mesh.reshape(np.prod(og_shape),3))
 		return pcs_mesh.reshape(*og_shape)
 
+	def pre_mesh(self, mesh, gamma=2*np.pi*42.576E6, rtype='r2', 
+		dsa=True, sbm=True):
+		"""
+		Calculate a PRE value at each location of cubic grid of points
+
+		Parameters
+		----------
+		mesh : array with shape (n,n,n,3)
+			a cubic grid as generated by the method <make_mesh>
+		gamma : float
+			the gyromagnetic ratio of the spin
+		rtype : str
+			either 'r1' or 'r2', the relaxation type
+		dsa : bool (optional)
+			when True (defualt), DSA or Curie spin relaxation is included
+		sbm : bool (optional)
+			when True (defualt), SBM spin relaxation is included
+
+		Returns
+		-------
+		pre_mesh : array with shape (n,n,n,1)
+			The same grid shape, with PRE values at the respective locations
+		"""
+		og_shape = mesh.shape[:3]
+		flat = mesh.reshape(np.prod(og_shape),3)
+		gamarr = np.ones(len(flat))*gamma
+		pre_mesh = self.fast_pre(flat, gamarr, rtype=rtype, dsa=dsa, sbm=sbm)
+		return pre_mesh.reshape(*og_shape)
+
 	def write_pymol_script(self, isoval=1.0, surfaceName='isomap', 
 		scriptName='isomap.pml', meshName='./isomap.pml.ccp4', pdbFile=None):
+		"""
+		Write a PyMol script to file which allows loading of the 
+		isosurface file
+
+		Parameters
+		----------
+		isoval : float (optional)
+			the contour level of the isosurface
+		surfaceName : str (optional)
+			the name of the isosurface file within PyMol
+		scriptName : str (optional)
+			the name of the PyMol script to load the tensor isosurface
+		meshName : str (optional)
+			the name of the binary isosurface file
+		pdbFile : str (optional)
+			if not <None>, the file name of the PDB file to be loaded with
+			the isosurface.
+		"""
 		posname = "pos_{}".format(surfaceName)
 		negname = "neg_{}".format(surfaceName)
 		oriname = "ori_{}".format(surfaceName)
@@ -1678,24 +1909,38 @@ class Metal(object):
 		s += "color pink, {}\n".format(oriname)
 		if pdbFile:
 			protName = ntpath.basename(pdbFile).replace('.pdb','')
-			s += "cmd.load(os.path.join(curdir, '{}'),'{}')\n".format(pdbFile, protName)
+			s += "cmd.load(os.path.join(curdir, '{}'),'{}')\n".format(
+				pdbFile, protName)
 			s += "show_as cartoon, {}\n".format(protName)
 		with open(scriptName, 'w') as o:
 			o.write(s)
 			print("{} script written".format(scriptName))
 
 	def write_isomap(self, mesh, bounds, fileName='isomap.pml.ccp4'):
+		"""
+		Write a PyMol script to file which allows loading of the 
+		isosurface file
+
+		Parameters
+		----------
+		mesh : 3D scalar np.ndarray of floats
+			the scalar field of PCS or PRE values in a cubic grid
+		bounds : tuple (origin, low, high, points)
+			as generated by :meth:`paramagpy.metal.Metal.make_mesh`
+		fileName : str (optional)
+			the filename of the isosurface file
+		"""
 		mesh = np.asarray(mesh, np.float32)
 		origin, low, high, points = bounds
 		with open(fileName, 'wb') as o:
 			for dim in points:
-				o.write(struct.pack('i', dim)) # number points / dim
+				o.write(struct.pack('i', dim)) # number points per dim
 			o.write(struct.pack('i',2)) # mode 2: 32bit float
 			for start in origin:
 				o.write(struct.pack('i', start)) # start point of map
 			for dim in points:
 				intervals = dim - 1
-				o.write(struct.pack('i', intervals)) # number intervals / dim
+				o.write(struct.pack('i', intervals)) # number intervals per dim
 			for scale in (high - low):
 				o.write(struct.pack('f', scale)) # cell dim scales
 			for i in range(3):
@@ -1722,10 +1967,6 @@ class Metal(object):
 			o.write(self.info(comment=False))
 
 
-
-
-
-
 def make_tensor(x, y, z, axial, rhombic, 
 	alpha, beta, gamma, lanthanide=None, temperature=298.15):
 	"""
@@ -1745,7 +1986,7 @@ def make_tensor(x, y, z, axial, rhombic,
 
 	Returns
 	-------
-	ChiTensor : object
+	ChiTensor : object :class:`paramagpy.metal.Metal`
 		a tensor object for calulating paramagnetic effects on 
 		nuclear spins in the pdb coordinate
 	"""
